/*
 * Hibernate Search, full-text search for your domain model
 *
 * License: GNU Lesser General Public License (LGPL), version 2.1 or later
 * See the lgpl.txt file in the root directory or <http://www.gnu.org/licenses/lgpl-2.1.html>.
 */
package org.hibernate.search.integrationtest.backend.elasticsearch.analysis;

import java.util.function.Consumer;

import org.hibernate.search.backend.elasticsearch.analysis.ElasticsearchAnalysisConfigurer;
import org.hibernate.search.backend.elasticsearch.analysis.model.dsl.ElasticsearchAnalysisDefinitionContainerContext;
import org.hibernate.search.backend.elasticsearch.cfg.ElasticsearchBackendSettings;
import org.hibernate.search.engine.mapper.mapping.building.spi.IndexModelBindingContext;
import org.hibernate.search.integrationtest.backend.tck.testsupport.util.rule.SearchSetupHelper;
import org.hibernate.search.util.SearchException;
import org.hibernate.search.util.impl.integrationtest.common.FailureReportUtils;
import org.hibernate.search.util.impl.test.SubTest;

import org.junit.Rule;
import org.junit.Test;

public class ElasticsearchAnalysisConfigurerIT {

	private static final String MAPPED_TYPE_NAME = "TypeName";
	private static final String BACKEND_NAME = "BackendName";
	private static final String INDEX_NAME = "IndexName";

	private static final String ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX = "Error while applying analysis configuration";

	@Rule
	public SearchSetupHelper setupHelper = new SearchSetupHelper();

	@Test
	public void error_invalidReference() {
		SubTest.expectException(
				() -> setup( "foobar" )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Unable to convert configuration property 'backends." + BACKEND_NAME
										+ "." + ElasticsearchBackendSettings.ANALYSIS_CONFIGURER + "'",
								"'foobar'",
								"Unable to find " + ElasticsearchAnalysisConfigurer.class.getName() + " implementation class: foobar"
						)
						.build()
				);
	}

	@Test
	public void error_failingConfigurer() {
		SubTest.expectException(
				() -> setup( FailingConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								FailingConfigurer.FAILURE_MESSAGE
						)
						.build()
				);
	}

	public static class FailingConfigurer implements ElasticsearchAnalysisConfigurer {
		private static final String FAILURE_MESSAGE = "Simulated failure for " + FailingConfigurer.class.getName();
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			throw new SimulatedFailure( FAILURE_MESSAGE );
		}
	}

	private static class SimulatedFailure extends RuntimeException {
		SimulatedFailure(String message) {
			super( message );
		}
	}

	@Test
	public void error_analyzer_namingConflict() {
		SubTest.expectException(
				() -> setup( AnalyzerNamingConflictConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Multiple analyzer definitions with the same name",
								"'analyzerName'"
						)
						.build()
				);
	}

	public static class AnalyzerNamingConflictConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.analyzer( "analyzerName" ).custom().withTokenizer( "whitespace" );
			context.analyzer( "anotherAnalyzerName" ).custom().withTokenizer( "whitespace" );
			context.analyzer( "analyzerName" ).type( "someType" );
		}
	}

	@Test
	public void error_normalizer_namingConflict() {
		SubTest.expectException(
				() -> setup( NormalizerNamingConflictConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Multiple normalizer definitions with the same name",
								"'normalizerName'"
						)
						.build()
				);
	}

	public static class NormalizerNamingConflictConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.normalizer( "normalizerName" ).custom();
			context.normalizer( "anotherNormalizerName" ).custom();
			context.normalizer( "normalizerName" ).custom();
		}
	}

	@Test
	public void error_tokenizer_namingConflict() {
		SubTest.expectException(
				() -> setup( TokenizerNamingConflictConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Multiple tokenizer definitions with the same name",
								"'tokenizerName'"
						)
						.build()
				);
	}

	public static class TokenizerNamingConflictConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.tokenizer( "tokenizerName" ).type( "someType" );
			context.tokenizer( "tokenizerName" ).type( "someType" );
		}
	}

	@Test
	public void error_tokenizer_missingType() {
		SubTest.expectException(
				() -> setup( TokenizerMissingTypeConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Invalid tokenizer definition for name 'tokenizerName'",
								"Tokenizer definitions must at least define the tokenizer type"
						)
						.build()
				);
	}

	public static class TokenizerMissingTypeConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.tokenizer( "tokenizerName" );
		}
	}

	@Test
	public void error_charFilter_namingConflict() {
		SubTest.expectException(
				() -> setup( CharFilterNamingConflictConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Multiple char filter definitions with the same name",
								"'charFilterName'"
						)
						.build()
				);
	}

	public static class CharFilterNamingConflictConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.charFilter( "charFilterName" ).type( "someType" );
			context.charFilter( "charFilterName" ).type( "someType" );
		}
	}

	@Test
	public void error_charFilter_missingType() {
		SubTest.expectException(
				() -> setup( CharFilterMissingTypeConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Invalid char filter definition for name 'charFilterName'",
								"Char filter definitions must at least define the char filter type"
						)
						.build()
				);
	}

	public static class CharFilterMissingTypeConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.charFilter( "charFilterName" );
		}
	}

	@Test
	public void error_tokenFilter_namingConflict() {
		SubTest.expectException(
				() -> setup( TokenFilterNamingConflictConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Multiple token filter definitions with the same name",
								"'tokenFilterName'"
						)
						.build()
				);
	}

	public static class TokenFilterNamingConflictConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.tokenFilter( "tokenFilterName" ).type( "someType" );
			context.tokenFilter( "tokenFilterName" ).type( "someType" );
		}
	}

	@Test
	public void error_tokenFilter_missingType() {
		SubTest.expectException(
				() -> setup( TokenFilterMissingTypeConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Invalid token filter definition for name 'tokenFilterName'",
								"Token filter definitions must at least define the token filter type"
						)
						.build()
				);
	}

	public static class TokenFilterMissingTypeConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.tokenFilter( "tokenFilterName" );
		}
	}

	@Test
	public void error_parameter_namingConflict() {
		SubTest.expectException(
				() -> setup( ParameterNamingConflictConfigurer.class.getName() )
		)
				.assertThrown()
				.isInstanceOf( SearchException.class )
				.hasMessageMatching( FailureReportUtils.buildSingleContextFailureReportPattern()
						.typeContext( MAPPED_TYPE_NAME )
						.indexContext( INDEX_NAME )
						.backendContext( BACKEND_NAME )
						.failure(
								ANALYSIS_CONFIGURER_ERROR_MESSAGE_PREFIX,
								"Multiple parameters with the same name",
								"'parameterName'",
								"'\"value1\"'",
								"'\"value2\"'"
						)
						.build()
				);
	}

	public static class ParameterNamingConflictConfigurer implements ElasticsearchAnalysisConfigurer {
		@Override
		public void configure(ElasticsearchAnalysisDefinitionContainerContext context) {
			context.analyzer( "analyzerName" ).type( "someType" )
					.param( "parameterName", "value1" )
					.param( "anotherParameterName", "someValue" )
					.param( "parameterName", "value2" );
		}
	}

	private void setup(String analysisConfigurer) {
		setup( analysisConfigurer, c -> { } );
	}

	private void setup(String analysisConfigurer, Consumer<IndexModelBindingContext> mappingContributor) {
		setupHelper.withDefaultConfiguration( BACKEND_NAME )
				.withProperty(
						"backends." + BACKEND_NAME + "." + ElasticsearchBackendSettings.ANALYSIS_CONFIGURER,
						analysisConfigurer
				)
				.withIndex(
						MAPPED_TYPE_NAME, INDEX_NAME,
						mappingContributor,
						indexManager -> { }
				)
				.setup();
	}
}
