/*
 * Hibernate Search, full-text search for your domain model
 *
 * License: GNU Lesser General Public License (LGPL), version 2.1 or later
 * See the lgpl.txt file in the root directory or <http://www.gnu.org/licenses/lgpl-2.1.html>.
 */
package org.hibernate.search.integrationtest.mapper.pojo.mapping.definition;

import static org.hibernate.search.util.impl.integrationtest.common.assertion.SearchResultAssert.assertThat;

import java.lang.invoke.MethodHandles;
import java.util.Collections;
import java.util.Iterator;
import java.util.Optional;

import org.hibernate.search.engine.backend.types.converter.runtime.spi.ToDocumentIdentifierValueConvertContext;
import org.hibernate.search.engine.backend.types.converter.runtime.spi.ToDocumentIdentifierValueConvertContextImpl;
import org.hibernate.search.engine.backend.types.converter.spi.ToDocumentIdentifierValueConverter;
import org.hibernate.search.engine.search.SearchQuery;
import org.hibernate.search.integrationtest.mapper.pojo.testsupport.types.PropertyTypeDescriptor;
import org.hibernate.search.integrationtest.mapper.pojo.testsupport.types.expectations.DefaultIdentifierBridgeExpectations;
import org.hibernate.search.integrationtest.mapper.pojo.testsupport.util.rule.JavaBeanMappingSetupHelper;
import org.hibernate.search.mapper.javabean.JavaBeanMapping;
import org.hibernate.search.mapper.javabean.mapping.context.impl.JavaBeanMappingContext;
import org.hibernate.search.mapper.javabean.session.JavaBeanSearchManager;
import org.hibernate.search.mapper.pojo.mapping.impl.PojoReferenceImpl;
import org.hibernate.search.mapper.pojo.search.PojoReference;
import org.hibernate.search.util.impl.integrationtest.common.rule.BackendMock;
import org.hibernate.search.util.impl.integrationtest.common.rule.StubSearchWorkBehavior;
import org.hibernate.search.util.impl.integrationtest.common.stub.backend.StubBackendUtils;
import org.hibernate.search.util.impl.integrationtest.common.stub.backend.document.model.StubIndexSchemaNode;
import org.hibernate.search.util.impl.test.SubTest;

import org.junit.Assume;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;

import org.assertj.core.api.Assertions;
import org.easymock.Capture;

/**
 * Test default identifier bridges for the {@code @DocumentId} annotation.
 */
@RunWith(Parameterized.class)
public class DocumentIdDefaultBridgeIT<I> {

	@Parameterized.Parameters(name = "{0}")
	public static Object[] types() {
		return PropertyTypeDescriptor.getAll().stream()
				.map( type -> new Object[] { type, type.getDefaultIdentifierBridgeExpectations() } )
				.toArray();
	}

	@Rule
	public BackendMock backendMock = new BackendMock( "stubBackend" );

	@Rule
	public JavaBeanMappingSetupHelper setupHelper = new JavaBeanMappingSetupHelper( MethodHandles.lookup() );

	private DefaultIdentifierBridgeExpectations<I> expectations;
	private JavaBeanMapping mapping;
	private StubIndexSchemaNode index1RootSchemaNode;
	private StubIndexSchemaNode index2RootSchemaNode;

	public DocumentIdDefaultBridgeIT(PropertyTypeDescriptor<I> typeDescriptor,
			Optional<DefaultIdentifierBridgeExpectations<I>> expectations) {
		Assume.assumeTrue(
				"Type " + typeDescriptor + " does not have a default identifier bridge", expectations.isPresent()
		);
		this.expectations = expectations.get();
	}

	@Before
	public void setup() {
		Capture<StubIndexSchemaNode> schemaCapture1 = Capture.newInstance();
		Capture<StubIndexSchemaNode> schemaCapture2 = Capture.newInstance();
		backendMock.expectSchema(
				DefaultIdentifierBridgeExpectations.TYPE_WITH_IDENTIFIER_BRIDGE_1_INDEX_NAME,
				b -> { },
				schemaCapture1
		);
		backendMock.expectSchema(
				DefaultIdentifierBridgeExpectations.TYPE_WITH_IDENTIFIER_BRIDGE_2_INDEX_NAME,
				b -> { },
				schemaCapture2
		);
		mapping = setupHelper.withBackendMock( backendMock )
				.setup( expectations.getTypeWithIdentifierBridge1(), expectations.getTypeWithIdentifierBridge2() );
		backendMock.verifyExpectationsMet();
		index1RootSchemaNode = schemaCapture1.getValue();
		index2RootSchemaNode = schemaCapture2.getValue();
	}

	@Test
	public void indexing() {
		try ( JavaBeanSearchManager manager = mapping.createSearchManager() ) {
			for ( I entityIdentifierValue : expectations.getEntityIdentifierValues() ) {
				Object entity = expectations.instantiateTypeWithIdentifierBridge1( entityIdentifierValue );
				manager.getMainWorkPlan().add( entity );
			}

			BackendMock.WorkCallListContext expectationSetter = backendMock.expectWorks(
					DefaultIdentifierBridgeExpectations.TYPE_WITH_IDENTIFIER_BRIDGE_1_INDEX_NAME
			);
			for ( String expectedDocumentIdentifierValue : expectations.getDocumentIdentifierValues() ) {
				expectationSetter.add(
						expectedDocumentIdentifierValue,
						b -> { }
				);
			}
			expectationSetter.preparedThenExecuted();
		}
		backendMock.verifyExpectationsMet();
	}

	@Test
	public void projection() {
		try ( JavaBeanSearchManager manager = mapping.createSearchManager() ) {
			Iterator<I> entityIdentifierIterator = expectations.getEntityIdentifierValues().iterator();
			for ( String documentIdentifierValue : expectations.getDocumentIdentifierValues() ) {
				I entityIdentifierValue = entityIdentifierIterator.next();
				backendMock.expectSearchReferences(
						Collections.singletonList(
								DefaultIdentifierBridgeExpectations.TYPE_WITH_IDENTIFIER_BRIDGE_1_INDEX_NAME ),
						b -> {
						},
						StubSearchWorkBehavior.of(
								1L,
								StubBackendUtils.reference(
										DefaultIdentifierBridgeExpectations.TYPE_WITH_IDENTIFIER_BRIDGE_1_INDEX_NAME,
										documentIdentifierValue
								)
						)
				);

				SearchQuery<PojoReference> query = manager.search( expectations.getTypeWithIdentifierBridge1() )
						.query()
						.asReference()
						.predicate( f -> f.matchAll() )
						.build();

				assertThat( query )
						.hasHitsExactOrder( new PojoReferenceImpl(
								expectations.getTypeWithIdentifierBridge1(),
								entityIdentifierValue
						) );
			}
			backendMock.verifyExpectationsMet();
		}
	}

	// Test behavior that backends expect from our bridges when using the DSLs
	@Test
	public void dslToIndexConverter() {
		// This cast may be unsafe, but only if something is deeply wrong, and then an exception will be thrown below
		@SuppressWarnings("unchecked")
		ToDocumentIdentifierValueConverter<I> dslToIndexConverter =
				(ToDocumentIdentifierValueConverter<I>) index1RootSchemaNode.getIdDslConverter();
		ToDocumentIdentifierValueConverter<?> compatibleDslToIndexConverter =
				index2RootSchemaNode.getIdDslConverter();
		ToDocumentIdentifierValueConvertContextImpl convertContext =
				new ToDocumentIdentifierValueConvertContextImpl( new JavaBeanMappingContext() );

		// isCompatibleWith must return true when appropriate
		Assertions.assertThat( dslToIndexConverter.isCompatibleWith( dslToIndexConverter ) ).isTrue();
		Assertions.assertThat( dslToIndexConverter.isCompatibleWith( compatibleDslToIndexConverter ) ).isTrue();
		Assertions.assertThat( dslToIndexConverter.isCompatibleWith( new IncompatibleToDocumentIdentifierValueConverter() ) )
				.isFalse();

		// convert and convertUnknown must behave appropriately on valid input
		Iterator<String> documentIdentifierIterator = expectations.getDocumentIdentifierValues().iterator();
		for ( I entityIdentifierValue : expectations.getEntityIdentifierValues() ) {
			String documentIdentifierValue = documentIdentifierIterator.next();
			Assertions.assertThat(
					dslToIndexConverter.convert( entityIdentifierValue, convertContext )
			)
					.isEqualTo( documentIdentifierValue );
			Assertions.assertThat(
					dslToIndexConverter.convertUnknown( entityIdentifierValue, convertContext )
			)
					.isEqualTo( documentIdentifierValue );
		}

		// convertUnknown must throw a runtime exception on invalid input
		SubTest.expectException(
				"convertUnknown on invalid input",
				() -> dslToIndexConverter.convertUnknown( new Object(), convertContext )
		)
				.assertThrown()
				.isInstanceOf( RuntimeException.class );
	}

	private static class IncompatibleToDocumentIdentifierValueConverter
			implements ToDocumentIdentifierValueConverter<Object> {
		@Override
		public String convert(Object value, ToDocumentIdentifierValueConvertContext context) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String convertUnknown(Object value, ToDocumentIdentifierValueConvertContext context) {
			throw new UnsupportedOperationException();
		}

		@Override
		public boolean isCompatibleWith(ToDocumentIdentifierValueConverter<?> other) {
			throw new UnsupportedOperationException();
		}
	}
}
